#include <QPixmapCache>
#include <QPixmap>
#include <QImage>
#include <QMatrix>
#include <QByteArray>

#include "flyer.h"
#include "options.h"
#include "gamelist.h"
#include "qmc2main.h"
#include "macros.h"

// external global variables
extern MainWindow *qmc2MainWindow;
extern Options *qmc2Options;
extern Gamelist *qmc2Gamelist;
extern bool qmc2UseFlyerFile;
extern bool qmc2GuiReady;
extern bool qmc2ReloadActive;
extern bool qmc2ScaledFlyer;
extern bool qmc2SmoothScaling;
extern QTreeWidgetItem *qmc2CurrentItem;
extern QSettings *qmc2Config;

QPixmap *currentFlyerPixmap;

Flyer::Flyer(QWidget *parent)
  : QFrame(parent)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Flyer::Flyer(QWidget *parent = 0x" + QString::number((ulong)parent, 16) + ")");
#endif

  setToolTip(tr("Game flyer image"));
  setStatusTip(tr("Game flyer image"));
  flyerFile = NULL;
  if ( qmc2UseFlyerFile ) {
    flyerFile = unzOpen((const char *)qmc2Config->value("Frontend/FilesAndDirectories/FlyerFile").toString().toAscii());
    if ( flyerFile == NULL )
      qmc2MainWindow->log(LOG_QMC2, tr("FATAL: can't open flyer file, please check access permissions for %1").arg(qmc2Config->value("Frontend/FilesAndDirectories/FlyerFile").toString()));
  }
}

Flyer::~Flyer()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Flyer::~Flyer()");
#endif

  if ( qmc2UseFlyerFile )
    unzClose(flyerFile);
}

void Flyer::paintEvent(QPaintEvent *e)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Flyer::paintEvent(QPaintEvent *e = 0x" + QString::number((ulong)e, 16) + ")");
#endif

  if ( !qmc2CurrentItem )
    return;

  if ( qmc2CurrentItem->text(0) == tr("Waiting for data...") )
    return;

  QTreeWidgetItem *topLevelItem = qmc2CurrentItem;
  while ( topLevelItem->parent() )
    topLevelItem = topLevelItem->parent();

  QPainter p(this);

  QString gameName = topLevelItem->child(0)->text(1);
  static QPixmap cachedPixmap;

  if ( QPixmapCache::find("fly_" + gameName, cachedPixmap) ) {
    currentFlyerPixmap = &cachedPixmap;
  } else {
    qmc2CurrentItem = topLevelItem;
    loadFlyer(gameName);
  }

  if ( qmc2ScaledFlyer )
    drawScaledImage(currentFlyerPixmap, &p);
  else
    drawCenteredImage(currentFlyerPixmap, &p);
}

bool Flyer::loadFlyer(QString gameName, bool checkOnly, QString *fileName)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, QString("DEBUG: Flyer::loadFlyer(QString gameName = %1, bool checkOnly = %2, QString *fileName = %3)").arg(gameName).arg(checkOnly).arg((qulonglong)fileName));
#endif

  static QPixmap pm;
  static char imageBuffer[QMC2_ZIP_BUFFER_SIZE];

  if ( fileName )
    *fileName = "";

  bool fileOk = TRUE;

  if ( qmc2UseFlyerFile ) {
    // use flyer file
    QByteArray imageData;
    int len, i;
    QString gameFile = gameName + ".png";

    if ( fileName )
      *fileName = gameFile;

    if ( unzLocateFile(flyerFile, (const char *)gameFile.toAscii(), 0) == UNZ_OK ) {
      if ( unzOpenCurrentFile(flyerFile) == UNZ_OK ) {
        while ( (len = unzReadCurrentFile(flyerFile, &imageBuffer, QMC2_ZIP_BUFFER_SIZE)) > 0 ) {
          for (i = 0; i < len; i++)
            imageData += imageBuffer[i];
        }
        unzCloseCurrentFile(flyerFile);
      } else
        fileOk = FALSE;
    } else
      fileOk = FALSE;

    if ( fileOk )
      fileOk = pm.loadFromData(imageData);

    if ( !checkOnly ) {
      if ( fileOk ) {
        QPixmapCache::insert("fly_" + gameName, pm); 
        currentFlyerPixmap = &pm;
      } else
        currentFlyerPixmap = &qmc2MainWindow->qmc2GhostImagePixmap;
    }
  } else {
    // use flyer directory
    QString imagePath = qmc2Config->value("Frontend/FilesAndDirectories/FlyerDirectory").toString() + gameName + ".png";

    if ( fileName )
      *fileName = gameName + ".png";

    if ( checkOnly ) {
      fileOk = pm.load(imagePath);
    } else {
      if ( pm.load(imagePath) ) {
        QPixmapCache::insert("fly_" + gameName, pm); 
        currentFlyerPixmap = &pm;
        fileOk = TRUE;
      } else {
        currentFlyerPixmap = &qmc2MainWindow->qmc2GhostImagePixmap;
        fileOk = FALSE;
      }
    }
  }

  return fileOk;
}

void Flyer::drawCenteredImage(QPixmap *pm, QPainter *p)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Flyer::drawCenteredImage(QPixmap *pm = 0x" + QString::number((ulong)pm, 16) + ", QPainter *p = 0x" + QString::number((ulong)p, 16) + ")");
#endif

  p->eraseRect(frameRect());

  if ( pm == NULL ) {
    p->end();
    return;
  }

  int posx = (frameRect().width() - pm->width()) / 2;
  int posy = (frameRect().height() - pm->height()) / 2;

  p->drawPixmap(posx, posy, *pm);
  p->end();
}

void Flyer::drawScaledImage(QPixmap *pm, QPainter *p)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Flyer::drawScaledImage(QPixmap *pm = 0x" + QString::number((ulong)pm, 16) + ", QPainter *p = 0x" + QString::number((ulong)p, 16) + ")");
#endif

  if ( pm == NULL ) {
    p->eraseRect(frameRect());
    p->end();
    return;
  }

  double desired_width;
  double desired_height;

  if ( pm->width() > pm->height() ) {
    desired_width  = contentsRect().width();
    desired_height = (double)pm->height() * (desired_width / (double)pm->width());
    if ( desired_height > contentsRect().height() ) {
      desired_height = contentsRect().height();
      desired_width  = (double)pm->width() * (desired_height / (double)pm->height());
    }
  } else {
    desired_height = contentsRect().height();
    desired_width  = (double)pm->width() * (desired_height / (double)pm->height());
    if ( desired_width > contentsRect().width() ) {
      desired_width = contentsRect().width();
      desired_height = (double)pm->height() * (desired_width / (double)pm->width());
    }
  }

  QPixmap pmScaled;

  if ( qmc2SmoothScaling )
    pmScaled = pm->scaled((int)desired_width, (int)desired_height, Qt::KeepAspectRatio, Qt::SmoothTransformation);
  else
    pmScaled = pm->scaled((int)desired_width, (int)desired_height, Qt::KeepAspectRatio, Qt::FastTransformation);

  drawCenteredImage(&pmScaled, p);
}
