#include <QFile>
#include <QTextStream>
#include <QHeaderView>
#include <QComboBox>
#include <QDoubleSpinBox>
#include <QSpinBox>
#include <QLineEdit>

#include "mameopt.h"
#include "options.h"
#include "qmc2main.h"

// external global variables
extern MainWindow *qmc2MainWindow;
extern Options *qmc2Options;
extern QSettings *qmc2Config;
extern MameOptions *qmc2GlobalMAMEOptions;
extern bool qmc2GuiReady;

QMap<QString, QList<MameOption> > MameOptions::templateMap;

QString optionDescription = "";
int optionType = MAMEOPT_TYPE_UNKNOWN;
MameOptions *mameOptions = NULL;

MameOptionDelegate::MameOptionDelegate(QObject *parent)
  : QItemDelegate(parent)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, QString("DEBUG: MameOptionDelegate::MameOptionDelegate(QObject *parent = %1)").arg((qulonglong)parent));
#endif

}

QWidget *MameOptionDelegate::createEditor(QWidget *parent, const QStyleOptionViewItem &option, const QModelIndex &index) const
{
  #define _MIN (-2 * 1024 * 1024)
  #define _MAX (2 * 1024 * 1024)

#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, QString("DEBUG: MameOptionDelegate::createEditor(QWidget *parent = %1, const QStyleOptionViewItem &option, const QModelIndex &index)").arg((qulonglong)parent));
#endif

  switch ( optionType ) {
    case MAMEOPT_TYPE_BOOL: {
      QComboBox *comboBoxEditor = new QComboBox(parent);
      comboBoxEditor->addItem(tr("false"));
      comboBoxEditor->addItem(tr("true"));
      comboBoxEditor->installEventFilter(const_cast<MameOptionDelegate*>(this));
      comboBoxEditor->setAccessibleName("comboBoxEditor");
      if ( !optionDescription.isEmpty() )
        comboBoxEditor->setToolTip(optionDescription);
      return comboBoxEditor;
    }

    case MAMEOPT_TYPE_INT: {
      QSpinBox *spinBoxEditor = new QSpinBox(parent);
      spinBoxEditor->setRange(_MIN, _MAX);
      spinBoxEditor->setSingleStep(1);
      spinBoxEditor->installEventFilter(const_cast<MameOptionDelegate*>(this));
      spinBoxEditor->setAccessibleName("spinBoxEditor");
      if ( !optionDescription.isEmpty() )
        spinBoxEditor->setToolTip(optionDescription);
      return spinBoxEditor;
    }

    case MAMEOPT_TYPE_FLOAT: {
      QDoubleSpinBox *doubleSpinBoxEditor = new QDoubleSpinBox(parent);
      doubleSpinBoxEditor->setRange(_MIN, _MAX);
      doubleSpinBoxEditor->setSingleStep(0.1);
      doubleSpinBoxEditor->setDecimals(6);
      doubleSpinBoxEditor->installEventFilter(const_cast<MameOptionDelegate*>(this));
      doubleSpinBoxEditor->setAccessibleName("doubleSpinBoxEditor");
      if ( !optionDescription.isEmpty() )
        doubleSpinBoxEditor->setToolTip(optionDescription);
      return doubleSpinBoxEditor;
    }

    case MAMEOPT_TYPE_STRING:
    default: {
      QLineEdit *lineEditEditor = new QLineEdit(parent);
      lineEditEditor->installEventFilter(const_cast<MameOptionDelegate*>(this));
      lineEditEditor->setAccessibleName("lineEditEditor");
      if ( !optionDescription.isEmpty() )
        lineEditEditor->setToolTip(optionDescription);
      return lineEditEditor;
    }
  }
}

void MameOptionDelegate::setEditorData(QWidget *editor, const QModelIndex &index) const
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, QString("DEBUG: MameOptionDelegate::setEditorData(QWidget *editor = %1, const QModelIndex &index)").arg((qulonglong)editor));
#endif

  if ( editor->accessibleName() == "comboBoxEditor" ) {
    QString value = index.model()->data(index, Qt::EditRole).toString();
    QComboBox *comboBox = static_cast<QComboBox *>(editor);
    comboBox->setCurrentIndex(comboBox->findText(value));
    if ( mameOptions )
      connect(comboBox, SIGNAL(activated(const QString &)), mameOptions, SLOT(optionChanged()));
  } else if ( editor->accessibleName() == "spinBoxEditor" ) {
    int value = index.model()->data(index, Qt::EditRole).toInt();
    QSpinBox *spinBox = static_cast<QSpinBox *>(editor);
    spinBox->setValue(value);
    if ( mameOptions )
      connect(spinBox, SIGNAL(valueChanged(int)), mameOptions, SLOT(optionChanged()));
  } else if ( editor->accessibleName() == "doubleSpinBoxEditor" ) {
    double value = index.model()->data(index, Qt::EditRole).toDouble();
    QDoubleSpinBox *doubleSpinBox = static_cast<QDoubleSpinBox *>(editor);
    doubleSpinBox->setValue(value);
    if ( mameOptions )
      connect(doubleSpinBox, SIGNAL(valueChanged(double)), mameOptions, SLOT(optionChanged()));
  } else {
    QString value = index.model()->data(index, Qt::EditRole).toString();
    QLineEdit *lineEdit = static_cast<QLineEdit *>(editor);
    lineEdit->setText(value);
    if ( mameOptions )
      connect(lineEdit, SIGNAL(textChanged(const QString &)), mameOptions, SLOT(optionChanged()));
  }
}

void MameOptionDelegate::setModelData(QWidget *editor, QAbstractItemModel *model, const QModelIndex &index) const
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, QString("DEBUG: MameOptionDelegate::setModelData(QWidget *editor = %1, QAbstractItemModel *model = %2, const QModelIndex &index)").arg((qulonglong)editor).arg((qulonglong)model));
#endif

  if ( editor->accessibleName() == "comboBoxEditor" ) {
    optionType = MAMEOPT_TYPE_BOOL;
    QComboBox *comboBox = static_cast<QComboBox*>(editor);
    QString v = comboBox->currentText();
    model->setData(index, v);
  } else if ( editor->accessibleName() == "spinBoxEditor" ) {
    optionType = MAMEOPT_TYPE_INT;
    QSpinBox *spinBox = static_cast<QSpinBox*>(editor);
    spinBox->interpretText();
    int v = spinBox->value();
    model->setData(index, v);
  } else if ( editor->accessibleName() == "doubleSpinBoxEditor" ) {
    optionType = MAMEOPT_TYPE_FLOAT;
    QDoubleSpinBox *doubleSpinBox = static_cast<QDoubleSpinBox*>(editor);
    doubleSpinBox->interpretText();
    double v = doubleSpinBox->value();
    model->setData(index, v);
  } else {
    optionType = MAMEOPT_TYPE_STRING;
    QLineEdit *lineEdit = static_cast<QLineEdit*>(editor);
    QString v = lineEdit->text();
    model->setData(index, v);
  }
}

void MameOptionDelegate::updateEditorGeometry(QWidget *editor, const QStyleOptionViewItem &option, const QModelIndex &index) const
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, QString("DEBUG: MameOptionDelegate::updateEditorGeometry(QWidget *editor = %1, const QStyleOptionViewItem &option, const QModelIndex &index)").arg((qulonglong)editor));
#endif

  editor->setGeometry(option.rect);
}

MameOptions::MameOptions(QString group, QWidget *parent)
  : QTreeWidget(parent)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: MameOptions::MameOptions(QString group = \"" + group + "\", QWidget *parent = 0x" + QString::number((ulong)parent, 16) + ")");
#endif

  // initialize
  connect(&searchTimer, SIGNAL(timeout()), this, SLOT(searchTimeout()));
  lineEditSearch = NULL;
  if ( !group.contains("Global") ) {
    mameOptions = NULL;
    setToolTip(tr("Game specific MAME configuration"));
    setStatusTip(tr("Game specific MAME configuration"));
  } else {
    mameOptions = this;
    setToolTip(tr("Global MAME configuration"));
    setStatusTip(tr("Global MAME configuration"));
  }
  changed = loadActive = FALSE;
  settingsGroup = group;
  setItemDelegate(&delegate);
  setAlternatingRowColors(TRUE);
  setColumnCount(2);  
  headerItem()->setText(0, tr("Option / Attribute"));
  headerItem()->setText(1, tr("Value"));
  header()->setClickable(FALSE);
  header()->setMovable(FALSE);
  header()->resizeSections(QHeaderView::Custom);
  header()->resizeSection(0, qmc2Config->value(settingsGroup + "/OptionColumnWidth", 200).toInt());
  if ( templateMap.count() == 0 )
    createTemplateMap();
  createMap();
}

MameOptions::~MameOptions()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: MameOptions::~MameOptions()");
#endif

  if ( header()->sectionSize(0) != 200 )
    qmc2Config->setValue(settingsGroup + "/OptionColumnWidth", header()->sectionSize(0));
  else
    qmc2Config->remove(settingsGroup + "/OptionColumnWidth");
}

void MameOptions::optionChanged()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: MameOptions::optionChanged()");
#endif

  changed = TRUE;
}

QString MameOptions::value(QString element, QString attribute)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: MameOptions::value(QString element = " + element + ", QString attribute = \"" + attribute + "\")");
#endif

  if ( element.contains(attribute) ) {
    QString valueString = element.remove(0, element.indexOf(attribute) + attribute.length() + 2);
    valueString = valueString.remove(valueString.indexOf("\""), valueString.lastIndexOf(">"));
    return valueString;
  }
  else {
    return QString::null;
  }
}

void MameOptions::load(bool overwrite)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: MameOptions::load(bool overwrite = " + QString(overwrite ? "TRUE" : "FALSE") + ")");
#endif

  loadActive = TRUE;
  qmc2Config->beginGroup(settingsGroup);
  QString sectionTitle;
  foreach ( sectionTitle, optionsMap.keys() ) {
    MameOption option;
    bool ok;
    int i;
    for (i = 0; i < optionsMap[sectionTitle].count(); i++) {
      option = optionsMap[sectionTitle][i];
      switch ( option.type ) {
        case MAMEOPT_TYPE_INT: {
          optionType = MAMEOPT_TYPE_INT;
          int v;
          if ( qmc2GlobalMAMEOptions != this ) {
            if ( overwrite )
              v = qmc2GlobalMAMEOptions->optionsMap[sectionTitle][i].value.toInt();
            else
              v = qmc2Config->value(option.name, qmc2GlobalMAMEOptions->optionsMap[sectionTitle][i].value.toInt()).toInt(&ok);
          } else
            v = qmc2Config->value(option.name, option.dvalue.toInt()).toInt(&ok);
          if ( ok ) {
            optionsMap[sectionTitle][i].item->setData(1, Qt::EditRole, v);
            optionsMap[sectionTitle][i].value.sprintf("%d", v);
            optionsMap[sectionTitle][i].valid = TRUE;
          }
          break;
        }

        case MAMEOPT_TYPE_FLOAT: {
          optionType = MAMEOPT_TYPE_FLOAT;
          double v;
          if ( qmc2GlobalMAMEOptions != this ) {
            if ( overwrite )
              v = qmc2GlobalMAMEOptions->optionsMap[sectionTitle][i].value.toDouble();
            else
              v = qmc2Config->value(option.name, qmc2GlobalMAMEOptions->optionsMap[sectionTitle][i].value.toDouble()).toDouble(&ok);
          } else
            v = qmc2Config->value(option.name, option.dvalue.toDouble()).toDouble(&ok);
          if ( ok ) {
            optionsMap[sectionTitle][i].item->setData(1, Qt::EditRole, v);
            optionsMap[sectionTitle][i].value.sprintf("%.6f", v);
            optionsMap[sectionTitle][i].valid = TRUE;
          }
          break;
        }

        case MAMEOPT_TYPE_BOOL: {
          optionType = MAMEOPT_TYPE_BOOL;
          QString v;
          if ( qmc2GlobalMAMEOptions != this ) {
            if ( overwrite )
              v = qmc2GlobalMAMEOptions->optionsMap[sectionTitle][i].value;
            else
              v = qmc2Config->value(option.name, qmc2GlobalMAMEOptions->optionsMap[sectionTitle][i].value).toString();
          } else
            v = qmc2Config->value(option.name, option.dvalue).toString();
          if ( !v.isEmpty() ) {
            optionsMap[sectionTitle][i].item->setData(1, Qt::EditRole, v);
            optionsMap[sectionTitle][i].value = v;
            optionsMap[sectionTitle][i].valid = TRUE;
          }
          break;
        }

        case MAMEOPT_TYPE_STRING:
        default: {
          optionType = MAMEOPT_TYPE_STRING;
          QString v;
          if ( qmc2GlobalMAMEOptions != this ) {
            if ( overwrite )
              v = qmc2GlobalMAMEOptions->optionsMap[sectionTitle][i].value;
            else
              v = qmc2Config->value(option.name, qmc2GlobalMAMEOptions->optionsMap[sectionTitle][i].value).toString();
          } else
            v = qmc2Config->value(option.name, option.dvalue).toString();
          if ( !v.isEmpty() ) {
            optionsMap[sectionTitle][i].item->setData(1, Qt::EditRole, v);
            optionsMap[sectionTitle][i].value = v;
            optionsMap[sectionTitle][i].valid = TRUE;
          }
          break;
        }
      }
    }
  }
  qmc2Config->endGroup();
  loadActive = changed = FALSE;
}

void MameOptions::save()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: MameOptions::save()");
#endif

  if ( loadActive )
    return;

  qmc2Config->beginGroup(settingsGroup);
  QString sectionTitle;
  foreach ( sectionTitle, optionsMap.keys() ) {
    QString vs;
    int i;
    for (i = 0; i < optionsMap[sectionTitle].count(); i++) {
      switch ( optionsMap[sectionTitle][i].type ) {
        case MAMEOPT_TYPE_INT: {
          int v = optionsMap[sectionTitle][i].item->data(1, Qt::EditRole).toInt();
          int gv = qmc2GlobalMAMEOptions->optionsMap[sectionTitle][i].item->data(1, Qt::EditRole).toInt();
          vs.sprintf("%d", v);
          if ( qmc2GlobalMAMEOptions == this ) {
            if ( v != optionsMap[sectionTitle][i].dvalue.toInt() ) {
              optionsMap[sectionTitle][i].value = vs;
              qmc2Config->setValue(optionsMap[sectionTitle][i].name, vs);
            } else {
              qmc2Config->remove(optionsMap[sectionTitle][i].name);
            }
          } else if ( v != gv && optionsMap[sectionTitle][i].valid ) {
            optionsMap[sectionTitle][i].value = vs;
            qmc2Config->setValue(optionsMap[sectionTitle][i].name, vs);
          } else {
            qmc2Config->remove(optionsMap[sectionTitle][i].name);
          }
          break;
        }

        case MAMEOPT_TYPE_FLOAT: {
          double v = optionsMap[sectionTitle][i].item->data(1, Qt::EditRole).toDouble();
          double gv = qmc2GlobalMAMEOptions->optionsMap[sectionTitle][i].item->data(1, Qt::EditRole).toDouble();
          vs.sprintf("%.6f", v);
          if ( qmc2GlobalMAMEOptions == this ) {
            if ( v != optionsMap[sectionTitle][i].dvalue.toDouble() ) {
              optionsMap[sectionTitle][i].value = vs;
              qmc2Config->setValue(optionsMap[sectionTitle][i].name, vs);
            } else {
              qmc2Config->remove(optionsMap[sectionTitle][i].name);
            }
          } else if ( v != gv && optionsMap[sectionTitle][i].valid ) {
            optionsMap[sectionTitle][i].value = vs;
            qmc2Config->setValue(optionsMap[sectionTitle][i].name, vs);
          } else {
            qmc2Config->remove(optionsMap[sectionTitle][i].name);
          }
          break;
        }

        case MAMEOPT_TYPE_BOOL: {
          QString vs = optionsMap[sectionTitle][i].item->data(1, Qt::EditRole).toString();
          QString gv = qmc2GlobalMAMEOptions->optionsMap[sectionTitle][i].item->data(1, Qt::EditRole).toString();
          if ( qmc2GlobalMAMEOptions == this ) {
            if ( vs != optionsMap[sectionTitle][i].dvalue ) {
              optionsMap[sectionTitle][i].value = vs;
              qmc2Config->setValue(optionsMap[sectionTitle][i].name, vs);
            } else {
              qmc2Config->remove(optionsMap[sectionTitle][i].name);
            }
          } else if ( vs != gv && optionsMap[sectionTitle][i].valid ) {
            optionsMap[sectionTitle][i].value = vs;
            qmc2Config->setValue(optionsMap[sectionTitle][i].name, vs);
          } else {
            qmc2Config->remove(optionsMap[sectionTitle][i].name);
          }
          break;
        }

        case MAMEOPT_TYPE_STRING:
        default: {
          QString vs = optionsMap[sectionTitle][i].item->data(1, Qt::EditRole).toString();
          QString gv = qmc2GlobalMAMEOptions->optionsMap[sectionTitle][i].item->data(1, Qt::EditRole).toString();
          if ( qmc2GlobalMAMEOptions == this ) {
            if ( vs != optionsMap[sectionTitle][i].dvalue ) {
              optionsMap[sectionTitle][i].value = vs;
              qmc2Config->setValue(optionsMap[sectionTitle][i].name, vs);
            } else {
              qmc2Config->remove(optionsMap[sectionTitle][i].name);
            }
          } else if ( vs != gv && optionsMap[sectionTitle][i].valid ) {
            optionsMap[sectionTitle][i].value = vs;
            qmc2Config->setValue(optionsMap[sectionTitle][i].name, vs);
          } else {
            qmc2Config->remove(optionsMap[sectionTitle][i].name);
          }
          break;
        }
      }
    }
  }
  qmc2Config->endGroup();
  changed = FALSE;
}

void MameOptions::createMap()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: MameOptions::createMap()");
#endif

  optionsMap.clear();
  QString sectionTitle;
  foreach ( sectionTitle, templateMap.keys() ) {
    QTreeWidgetItem *sectionItem = new QTreeWidgetItem(this);
    sectionItem->setText(0, sectionTitle);
    optionsMap[sectionTitle] = templateMap[sectionTitle];
    int i;
    for (i = 0; i < optionsMap[sectionTitle].count(); i++ ) {
      optionsMap[sectionTitle][i].value = optionsMap[sectionTitle][i].dvalue;
      MameOption mameOption = optionsMap[sectionTitle].at(i);
      QTreeWidgetItem *optionItem = new QTreeWidgetItem(sectionItem);
      optionsMap[sectionTitle][i].item = optionItem;
      optionItem->setText(0, mameOption.name);
      optionItem->setText(1, mameOption.value);
      optionType = mameOption.type;
      QTreeWidgetItem *childItem;
      childItem = new QTreeWidgetItem(optionItem);
      childItem->setText(0, tr("Type"));
      switch ( mameOption.type ) {
        case MAMEOPT_TYPE_BOOL:
           childItem->setText(1, tr("bool"));
           break;

        case MAMEOPT_TYPE_INT:
           childItem->setText(1, tr("int"));
           break;

        case MAMEOPT_TYPE_FLOAT:
           childItem->setText(1, tr("float"));
           break;

        case MAMEOPT_TYPE_STRING:
           childItem->setText(1, tr("string"));
           break;

        default:
           childItem->setText(1, tr("unknown"));
           break;
      }
      if ( !mameOption.shortname.isEmpty() ) {
        childItem = new QTreeWidgetItem(optionItem);
        childItem->setText(0, tr("Short name"));
        childItem->setText(1, mameOption.shortname);
      }
      if ( !mameOption.dvalue.isEmpty() ) {
        childItem = new QTreeWidgetItem(optionItem);
        childItem->setText(0, tr("Default"));
        childItem->setText(1, mameOption.dvalue);
      }
      if ( !mameOption.description.isEmpty() ) {
        optionDescription = mameOption.description;
        childItem = new QTreeWidgetItem(optionItem);
        childItem->setText(0, tr("Description"));
        childItem->setText(1, mameOption.description);
      } else
        optionDescription = "";
      openPersistentEditor(optionItem, 1);
    }
  } 
}

void MameOptions::createTemplateMap()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: MameOptions::createTemplateMap()");
#endif

  qmc2MainWindow->log(LOG_QMC2, tr("creating template configuration map"));
  
  templateMap.clear();
  QString templateFile = qmc2Config->value("Frontend/FilesAndDirectories/OptionsTemplateFile").toString();
  if ( templateFile.isEmpty() )
    templateFile = "data/opt/template.xml";
  QFile qmc2TemplateFile(templateFile);
  QString lang = qmc2Config->value("Frontend/GUI/Language").toString();
  if ( lang.isEmpty() )
    lang = "us";
  if ( qmc2TemplateFile.open(QFile::ReadOnly) ) {
    QTextStream ts(&qmc2TemplateFile);
    QString s = ts.readAll();
    qmc2TemplateFile.close();
    QStringList sl = s.split("\n");
    int i;
    QString sectionName, sectionTitle;
    for (i = 0; i < sl.count(); i++) {
      s = sl[i].simplified();
      if ( !s.isEmpty() ) {
        if ( !s.contains("template") && !s.contains("<?xml") ) {
          if ( s.contains("<section ") ) {
            sectionTitle = value(s, "description_" + lang);
#ifdef QMC2_DEBUG
            qmc2MainWindow->log(LOG_QMC2, "DEBUG: MameOptions::createTemplateMap(): new section: description = \"" + sectionTitle + "\"");
#endif
            templateMap[sectionTitle].clear();
          } else if ( !s.contains("</section>") ) {
            MameOption opt(value(s, " name"), value(s, " shortname"), value(s, " type"), value(s, " default"), value(s, " description_" + lang), QString::null, NULL, FALSE);
            templateMap[sectionTitle].append(opt);
#ifdef QMC2_DEBUG
            qmc2MainWindow->log(LOG_QMC2, "DEBUG: MameOptions::createTemplateMap(): new option: name = \"" + opt.name + "\", shortname = \"" + opt.shortname + "\", type = \"" + QString::number(opt.type) + "\", default = \"" + opt.dvalue + "\", description = \"" + opt.description + "\", value = \"" + opt.value + "\"");
#endif
          }
        }
      }
    }
  } else {
    qmc2MainWindow->log(LOG_QMC2, tr("FATAL: can't open options template file"));
  }
}

void MameOptions::keyPressEvent(QKeyEvent *e)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: MameOptions::keyPressEvent(QKeyPressEvent *e)");
#endif

  if ( !e->text().isEmpty() ) {
    if ( !lineEditSearch )
      lineEditSearch = new QLineEdit(this);
    lineEditSearch->show();
    lineEditSearch->raise();
    lineEditSearch->event(e);
    searchTimer.start(QMC2_SEARCH_TIMEOUT);
    clearSelection();
    if ( !lineEditSearch->text().isEmpty() ) {
      int i;
      bool madeCurrent = FALSE;
      QList<QTreeWidgetItem *> foundItems = findItems(lineEditSearch->text(), Qt::MatchRecursive | Qt::MatchContains, 0);
      for (i = 0; i < foundItems.count(); i++) {
        QTreeWidgetItem *p = foundItems[i];
        if ( p->parent() ) {
          if ( !p->parent()->parent() ) {
            p->setSelected(TRUE);
            p->parent()->setSelected(TRUE);
            if ( !madeCurrent ) {
              scrollToItem(p);
              setCurrentItem(p);
              madeCurrent = TRUE;
            }
          }
        }
      }
    }
    e->accept();
  } else {
    QAbstractItemView::keyPressEvent(e);
  }
}

void MameOptions::searchTimeout()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: MameOptions::searchTimeout()");
#endif
  searchTimer.stop();
  lineEditSearch->close();
  delete lineEditSearch;
  lineEditSearch = NULL;
}
