#include <QPixmapCache>
#include <QTranslator>
#include <QFileInfo>
#include <QFileDialog>
#include <QFontDialog>
#include <QMessageBox>
#include <QTimer>
#include <QMap>
#include <QStyleFactory>
#include <QBitArray>

#include "options.h"
#include "mameopt.h"
#include "preview.h"
#include "flyer.h"
#include "qmc2main.h"
#include "gamelist.h"
#include "macros.h"
#include "unzip.h"

// external global variables
extern MainWindow *qmc2MainWindow;
extern Options *qmc2Options;
extern bool qmc2GuiReady;
extern bool qmc2CleaningUp;
extern bool qmc2StartingUp;
extern bool qmc2EarlyStartup;
extern bool qmc2ScaledPreview;
extern bool qmc2ScaledFlyer;
extern bool qmc2SmoothScaling;
extern bool qmc2VerifyActive;
extern bool qmc2ImageCheckActive;
extern bool qmc2SampleCheckActive;
extern bool qmc2UsePreviewFile;
extern bool qmc2UseFlyerFile;
extern bool qmc2UseIconFile;
extern int qmc2GamelistResponsiveness;
extern QTranslator *qmc2Translator;
extern QTranslator *qmc2QtTranslator;
extern MameOptions *qmc2GlobalMAMEOptions;
extern MameOptions *qmc2MAMEOptions;
extern Preview *qmc2Preview;
extern Flyer *qmc2Flyer;
extern Gamelist *qmc2Gamelist;
extern int qmc2SortCriteria;
extern Qt::SortOrder qmc2SortOrder;
extern QMap<QString, QString> qmc2GamelistNameMap;
extern QSettings *qmc2Config;
extern QBitArray qmc2Filter;
extern unzFile qmc2IconFile;

Options::Options(QWidget *parent)
  : QDialog(parent)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::Options(QWidget *parent = 0x" + QString::number((ulong)parent, 16) + ")");
#endif

  qmc2Filter.resize(QMC2_ROMSTATE_COUNT);
  QSettings::setPath(QSettings::IniFormat, QSettings::SystemScope, QString(XSTR(SYSCONFDIR)) + "/qmc2");
  QString userScopePath = QDir::homePath() + "/.qmc2";
  QSettings::setPath(QSettings::IniFormat, QSettings::UserScope, userScopePath);
  QDir userScopeDir(userScopePath);
  if ( !userScopeDir.exists() )
    userScopeDir.mkdir(userScopePath);
  config = new QSettings(QSettings::IniFormat, QSettings::UserScope, "qmc2");

  setupLanguage();
  setupUi(this);

  if ( !config->isWritable() ) {
    qmc2MainWindow->log(LOG_QMC2, tr("WARNING: configuration is not writeable, please check access permissions for ") + config->fileName());
  }

  restoreCurrentConfig();

  // restore layout
  if ( config->value("Frontend/GUI/RestoreLayout").toBool() ) {
    tabWidgetFrontendSettings->setCurrentIndex(config->value("Frontend/Layout/OptionsWidget/FrontendTab", 0).toInt());
    tabWidgetGlobalMAMESetup->setCurrentIndex(config->value("Frontend/Layout/OptionsWidget/MAMETab", 0).toInt());
    tabWidgetOptions->setCurrentIndex(config->value("Frontend/Layout/OptionsWidget/OptionsTab", 0).toInt());
    QStringList cl = config->allKeys();
    if ( cl.contains("Frontend/Layout/OptionsWidget/Size") )
      resize(config->value("Frontend/Layout/OptionsWidget/Size").toSize());
    if ( cl.contains("Frontend/Layout/OptionsWidget/Position") )
      move(config->value("Frontend/Layout/OptionsWidget/Position").toPoint());
    if ( cl.contains("Frontend/Layout/OptionsWidget/Visible") )
      if ( config->value("Frontend/Layout/OptionsWidget/Visible").toBool() )
        show();
  }

  QTimer::singleShot(0, this, SLOT(on_pushButtonApply_clicked()));
}

Options::~Options()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::~Options()");
#endif

  if ( config->value("Frontend/GUI/SaveLayout").toBool() ) {
    config->setValue("Frontend/Layout/OptionsWidget/FrontendTab", tabWidgetFrontendSettings->currentIndex());
    config->setValue("Frontend/Layout/OptionsWidget/MAMETab", tabWidgetGlobalMAMESetup->currentIndex());
    config->setValue("Frontend/Layout/OptionsWidget/OptionsTab", tabWidgetOptions->currentIndex());
  }

  delete config;
  close();
}

void Options::apply()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::apply()");
#endif
  qmc2MainWindow->statusBar()->setShown(config->value("Frontend/GUI/Statusbar").toBool());
  qmc2MainWindow->toolbar->setShown(config->value("Frontend/GUI/Toolbar").toBool());
  QFont f;
  f.fromString(config->value("Frontend/GUI/Font").toString());
  qApp->setFont(f);
  QFontMetrics fm(f);
  QSize iconSize(fm.height() - 3, fm.height() - 3);
  qmc2MainWindow->treeWidgetGamelist->setIconSize(iconSize);
  qmc2MainWindow->treeWidgetHierarchy->setIconSize(iconSize);
  pushButtonApply->setIconSize(iconSize);
  pushButtonRestore->setIconSize(iconSize);
  pushButtonDefault->setIconSize(iconSize);
  toolButtonBrowseFont->setIconSize(iconSize);
  toolButtonBrowseTemporaryFile->setIconSize(iconSize);
  toolButtonBrowseFrontendLogFile->setIconSize(iconSize);
  toolButtonBrowseFavoritesFile->setIconSize(iconSize);
  toolButtonBrowseHistoryFile->setIconSize(iconSize);
  toolButtonBrowseROMStateCacheFile->setIconSize(iconSize);
  toolButtonBrowseDataDirectory->setIconSize(iconSize);
  toolButtonBrowsePreviewDirectory->setIconSize(iconSize);
  toolButtonBrowsePreviewFile->setIconSize(iconSize);
  toolButtonBrowseFlyerDirectory->setIconSize(iconSize);
  toolButtonBrowseFlyerFile->setIconSize(iconSize);
  toolButtonBrowseIconDirectory->setIconSize(iconSize);
  toolButtonBrowseIconFile->setIconSize(iconSize);
  toolButtonShowC->setIconSize(iconSize);
  toolButtonShowM->setIconSize(iconSize);
  toolButtonShowI->setIconSize(iconSize);
  toolButtonShowN->setIconSize(iconSize);
  toolButtonShowU->setIconSize(iconSize);
  comboBoxSortOrder->setIconSize(iconSize);
  toolButtonBrowseExecutableFile->setIconSize(iconSize);
  toolButtonBrowseMAMELogFile->setIconSize(iconSize);
  toolButtonBrowseOptionsTemplateFile->setIconSize(iconSize);
}

void Options::setupLanguage()
{
  QString lang = config->value("Frontend/GUI/Language").toString();
  if ( qmc2Translator ) {
    qApp->removeTranslator(qmc2Translator); 
    delete qmc2Translator;
  }
  QString directory = config->value("Frontend/FilesAndDirectories/DataDirectory").toString() + "lng/";
  qmc2Translator = new QTranslator(0);
  qmc2Translator->load(directory + QString("qmc2_") + lang + ".qm");
  qApp->installTranslator(qmc2Translator);

  if ( qmc2QtTranslator ) {
    qApp->removeTranslator(qmc2QtTranslator);
    delete qmc2QtTranslator;
  }
  qmc2QtTranslator = new QTranslator(0);
  qmc2QtTranslator->load(directory + QString("qt_") + lang + ".qm");
  qApp->installTranslator(qmc2QtTranslator);

#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::setupLanguage(): loaded translations for language " + lang.toUpper());
#endif
}

void Options::on_pushButtonOk_clicked()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::on_pushButtonOk_clicked()");
#endif

  on_pushButtonApply_clicked();
}

void Options::on_pushButtonCancel_clicked()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::on_pushButtonCancel_clicked()");
#endif

  restoreCurrentConfig();
}

void Options::on_pushButtonRestore_clicked()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::on_pushButtonRestore_clicked()");
#endif

  restoreCurrentConfig();
}

void Options::on_pushButtonApply_clicked()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::on_pushButtonApply_clicked()");
#endif

  static int oldCacheSize = 0;
  QString s;
  int i;
  bool needRestart = FALSE,
       needResort = FALSE,
       needRecreateTemplateMap = FALSE,
       needFilter = FALSE,
       needReopenPreviewFile = FALSE,
       needReopenFlyerFile = FALSE,
       needReopenIconFile = FALSE;

  // General
#ifdef BETA
  config->setValue("Version", QString::number(MAJOR) + "." + QString::number(MINOR) + ".b" + QString::number(BETA));
#else 
  config->setValue("Version", QString::number(MAJOR) + "." + QString::number(MINOR));
#endif

  // Frontend

  // GUI
  config->setValue("Frontend/GUI/SaveLayout", checkBoxSaveLayout->isChecked());
  config->setValue("Frontend/GUI/RestoreLayout", checkBoxRestoreLayout->isChecked());
  config->setValue("Frontend/GUI/SaveGameSelection", checkBoxSaveGameSelection->isChecked());
  config->setValue("Frontend/GUI/RestoreGameSelection", checkBoxRestoreGameSelection->isChecked());
  config->setValue("Frontend/GUI/Statusbar", checkBoxStatusbar->isChecked());
  config->setValue("Frontend/GUI/Toolbar", checkBoxToolbar->isChecked());
  qmc2ScaledPreview = checkBoxScaledPreview->isChecked();
  config->setValue("Frontend/GUI/ScaledPreview", qmc2ScaledPreview);
  qmc2ScaledFlyer = checkBoxScaledFlyer->isChecked();
  config->setValue("Frontend/GUI/ScaledFlyer", qmc2ScaledFlyer);
  qmc2SmoothScaling = checkBoxSmoothScaling->isChecked();
  config->setValue("Frontend/GUI/SmoothScaling", qmc2SmoothScaling);
  s = comboBoxLanguage->currentText().left(2).toLower();
  needRestart = needRestart || (config->value("Frontend/GUI/Language").toString() != s);
  config->setValue("Frontend/GUI/Language", s);
  s = comboBoxStyle->currentText();
  if ( s == QObject::tr("Default") ) 
    s = "Default";
  config->setValue("Frontend/GUI/Style", s);
  qmc2MainWindow->setupStyle(s);
  config->setValue("Frontend/GUI/Font", lineEditFont->text());
  if ( spinBoxPixmapCacheSize->value() != oldCacheSize ) {
    oldCacheSize = spinBoxPixmapCacheSize->value();
    qmc2MainWindow->log(LOG_QMC2, tr("image cache size set to %1 MB").arg(oldCacheSize));
  }
  config->setValue("Frontend/GUI/PixmapCacheSize", oldCacheSize);
  QPixmapCache::setCacheLimit(oldCacheSize * 1024);
  config->setValue("Frontend/GUI/KillEmulatorsOnExit", checkBoxKillEmulatorsOnExit->isChecked());

  // Files and directories
  config->setValue("Frontend/FilesAndDirectories/TemporaryFile", lineEditTemporaryFile->text());
  config->setValue("Frontend/FilesAndDirectories/LogFile", lineEditFrontendLogFile->text());
  needReopenPreviewFile = (qmc2UsePreviewFile != (stackedWidgetPreview->currentIndex() == 1));
  needReopenPreviewFile |= (config->value("Frontend/FilesAndDirectories/PreviewFile") != lineEditPreviewFile->text());
  qmc2UsePreviewFile = (stackedWidgetPreview->currentIndex() == 1);
  config->setValue("Frontend/FilesAndDirectories/UsePreviewFile", qmc2UsePreviewFile);
  config->setValue("Frontend/FilesAndDirectories/PreviewDirectory", lineEditPreviewDirectory->text());
  config->setValue("Frontend/FilesAndDirectories/PreviewFile", lineEditPreviewFile->text());
  needReopenFlyerFile = (qmc2UseFlyerFile != (stackedWidgetFlyer->currentIndex() == 1));
  needReopenFlyerFile |= (config->value("Frontend/FilesAndDirectories/FlyerFile") != lineEditFlyerFile->text());
  qmc2UseFlyerFile = (stackedWidgetFlyer->currentIndex() == 1);
  config->setValue("Frontend/FilesAndDirectories/UseFlyerFile", qmc2UseFlyerFile);
  config->setValue("Frontend/FilesAndDirectories/FlyerDirectory", lineEditFlyerDirectory->text());
  config->setValue("Frontend/FilesAndDirectories/FlyerFile", lineEditFlyerFile->text());
  needReopenIconFile = (qmc2UseIconFile != (stackedWidgetIcon->currentIndex() == 1));
  needReopenIconFile |= (config->value("Frontend/FilesAndDirectories/IconFile") != lineEditIconFile->text());
  qmc2UseIconFile = (stackedWidgetIcon->currentIndex() == 1);
  config->setValue("Frontend/FilesAndDirectories/UseIconFile", qmc2UseIconFile);
  config->setValue("Frontend/FilesAndDirectories/IconDirectory", lineEditIconDirectory->text());
  config->setValue("Frontend/FilesAndDirectories/IconFile", lineEditIconFile->text());
  config->setValue("Frontend/FilesAndDirectories/DataDirectory", lineEditDataDirectory->text());
  s = lineEditOptionsTemplateFile->text();
  needRecreateTemplateMap = needRecreateTemplateMap || (config->value("Frontend/FilesAndDirectories/OptionsTemplateFile").toString() != s );
  config->setValue("Frontend/FilesAndDirectories/OptionsTemplateFile", s);
  config->setValue("Frontend/FilesAndDirectories/FavoritesFile", lineEditFavoritesFile->text());
  config->setValue("Frontend/FilesAndDirectories/HistoryFile", lineEditHistoryFile->text());
  config->setValue("Frontend/FilesAndDirectories/ROMStateCacheFile", lineEditROMStateCacheFile->text());

  // Gamelist
  config->setValue("Frontend/Gamelist/SortOnline", checkBoxSortOnline->isChecked());
  qmc2GamelistResponsiveness = spinBoxResponsiveness->value();
  config->setValue("Frontend/Gamelist/Responsiveness", qmc2GamelistResponsiveness);
  i = comboBoxSortCriteria->currentIndex();
  needResort = (i != qmc2SortCriteria);
  int oldSortCriteria = qmc2SortCriteria;
  qmc2SortCriteria = i;
  config->setValue("Frontend/Gamelist/SortCriteria", qmc2SortCriteria);
  i = comboBoxSortOrder->currentIndex();
  needResort = needResort || (i == 0 ? qmc2SortOrder != Qt::AscendingOrder : qmc2SortOrder != Qt::DescendingOrder);
  Qt::SortOrder oldSortOrder = qmc2SortOrder; 
  qmc2SortOrder = (i == 0 ? Qt::AscendingOrder : Qt::DescendingOrder);
  config->setValue("Frontend/Gamelist/SortOrder", qmc2SortOrder);
  QBitArray newFilter(QMC2_ROMSTATE_COUNT);
  config->setValue("Frontend/Gamelist/ShowC", toolButtonShowC->isChecked());
  newFilter.setBit(QMC2_ROMSTATE_INT_C, toolButtonShowC->isChecked());
  config->setValue("Frontend/Gamelist/ShowM", toolButtonShowM->isChecked());
  newFilter.setBit(QMC2_ROMSTATE_INT_M, toolButtonShowM->isChecked());
  config->setValue("Frontend/Gamelist/ShowI", toolButtonShowI->isChecked());
  newFilter.setBit(QMC2_ROMSTATE_INT_I, toolButtonShowI->isChecked());
  config->setValue("Frontend/Gamelist/ShowN", toolButtonShowN->isChecked());
  newFilter.setBit(QMC2_ROMSTATE_INT_N, toolButtonShowN->isChecked());
  config->setValue("Frontend/Gamelist/ShowU", toolButtonShowU->isChecked());
  newFilter.setBit(QMC2_ROMSTATE_INT_U, toolButtonShowU->isChecked());
  needFilter = (qmc2Filter != newFilter);
  qmc2Filter = newFilter;

  // MAME

  // Configuration
  if ( qmc2GuiReady ) {
    if ( qmc2GlobalMAMEOptions->changed ) {
      if ( qmc2MAMEOptions ) {
        switch ( QMessageBox::question(this, tr("Confirm"), tr("An open game-specific MAME configuration has been detected.\nUse local game-setting, overwrite with global setup or don't apply?"), tr("&Local"), tr("&Overwrite"), tr("Do&n't apply"), 0, 2) ) {
           case 0:
             qmc2MAMEOptions->save();
             qmc2GlobalMAMEOptions->save();
             qmc2GlobalMAMEOptions->load();
             qmc2MAMEOptions->load();
             break;

           case 1:
             qmc2GlobalMAMEOptions->save();
             qmc2GlobalMAMEOptions->load();
             qmc2MAMEOptions->load(TRUE);
             qmc2MAMEOptions->save();
             break;

           case 2: 
           default:
             break;
        }
      } else {
        qmc2GlobalMAMEOptions->save();
        qmc2GlobalMAMEOptions->load();
      }
    }
  }

  // Files and directories
  config->setValue("MAME/FilesAndDirectories/ExecutableFile", lineEditExecutableFile->text());
  config->setValue("MAME/FilesAndDirectories/LogFile", lineEditMAMELogFile->text());

  // sync settings (write settings to disk) and apply
  config->sync();
  applied = TRUE;
  if ( qmc2GuiReady )
    apply();

  if ( needRestart )
    qmc2MainWindow->log(LOG_QMC2, tr("please restart for some changes to take effect"));

  if ( needRecreateTemplateMap )
    qmc2MainWindow->on_actionRecreateTemplateMap_activated();

  if ( needResort ) {
    bool doResort = TRUE;

    if ( qmc2VerifyActive ) {
      qmc2MainWindow->log(LOG_QMC2, tr("re-sort of gamelist impossible at this time"));
      qmc2MainWindow->log(LOG_QMC2, tr("please wait for ROM verification to finish and try again"));
      qmc2SortCriteria = oldSortCriteria;
      qmc2SortOrder = oldSortOrder;
      doResort = FALSE;
    }

    if ( doResort ) {
      QString sortCirtieria = "?";
      switch ( qmc2SortCriteria ) {
        case QMC2_SORT_BY_DESCRIPTION:
          sortCirtieria = QObject::tr("game description");
          break;
        case QMC2_SORT_BY_ROM_STATE:
          sortCirtieria = QObject::tr("ROM state");
          break;
        case QMC2_SORT_BY_YEAR:
          sortCirtieria = QObject::tr("year");
          break;
        case QMC2_SORT_BY_MANUFACTURER:
          sortCirtieria = QObject::tr("manufacturer");
          break;
      }
      qmc2MainWindow->log(LOG_QMC2, tr("sorting gamelist by %1 in %2 order").arg(sortCirtieria).arg(qmc2SortOrder == Qt::AscendingOrder ? tr("ascending") : tr("descending")));
      qApp->processEvents();
      QList<QTreeWidgetItem *> itemList = qmc2MainWindow->treeWidgetGamelist->findItems("*", Qt::MatchContains | Qt::MatchWildcard);
      for (i = 0; i < itemList.count(); i++) {
        if ( itemList[i]->childCount() > 1 ) {
          qmc2MainWindow->treeWidgetGamelist->collapseItem(itemList[i]);
          QList<QTreeWidgetItem *> childrenList = itemList[i]->takeChildren();
          int j;
          for (j = 0; j < childrenList.count(); j++)
            delete childrenList[j];
          QTreeWidgetItem *nameItem = new QTreeWidgetItem(itemList[i]);
          nameItem->setText(0, tr("Waiting for data..."));
          nameItem->setText(1, qmc2GamelistNameMap[itemList[i]->text(0)]);
          qApp->processEvents();
        }
      }
      qApp->processEvents();
      qmc2MainWindow->treeWidgetGamelist->sortItems(0, qmc2SortOrder);
      qmc2MainWindow->treeWidgetHierarchy->sortItems(0, qmc2SortOrder);
      QTimer::singleShot(0, qmc2MainWindow, SLOT(scrollToCurrentItem()));
    }
  }

  if ( needFilter )
    qmc2Gamelist->filter();

  if ( qmc2Preview ) {
    if ( needReopenPreviewFile ) {
      if ( qmc2UsePreviewFile ) {
        qmc2Preview->previewFile = unzOpen((const char *)config->value("Frontend/FilesAndDirectories/PreviewFile").toString().toAscii());
        if ( qmc2Preview->previewFile == NULL )
          qmc2MainWindow->log(LOG_QMC2, tr("FATAL: can't open preview file, please check access permissions for %1").arg(config->value("Frontend/FilesAndDirectories/PreviewFile").toString()));
      } else
        unzClose(qmc2Preview->previewFile);
    }
    qmc2Preview->repaint();
  }

  if ( qmc2Flyer ) {
    if ( needReopenFlyerFile ) {
      if ( qmc2UseFlyerFile ) {
        qmc2Flyer->flyerFile = unzOpen((const char *)config->value("Frontend/FilesAndDirectories/FlyerFile").toString().toAscii());
        if ( qmc2Flyer->flyerFile == NULL )
          qmc2MainWindow->log(LOG_QMC2, tr("FATAL: can't open flyer file, please check access permissions for %1").arg(config->value("Frontend/FilesAndDirectories/FlyerFile").toString()));
      } else
        unzClose(qmc2Flyer->flyerFile);
    }
    qmc2Flyer->repaint();
  }

  if ( needReopenIconFile ) {
    if ( qmc2UseIconFile ) {
      qmc2IconFile = unzOpen((const char *)config->value("Frontend/FilesAndDirectories/IconFile").toString().toAscii());
      if ( qmc2IconFile == NULL )
        qmc2MainWindow->log(LOG_QMC2, tr("FATAL: can't open icon file, please check access permissions for %1").arg(config->value("Frontend/FilesAndDirectories/IconFile").toString()));
    } else
      unzClose(qmc2IconFile);
  }
}

void Options::on_pushButtonDefault_clicked()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::on_pushButtonDefault_clicked()");
#endif

  restoreCurrentConfig(TRUE);
}

void Options::restoreCurrentConfig(bool useDefaultSettings)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::restoreCurrentConfig()");
#endif

  if ( useDefaultSettings ) {
    QString fn = config->fileName();
    delete config;
    QFile f(fn);
    f.copy(fn + ".bak");
    f.remove();
    config = new QSettings(QSettings::IniFormat, QSettings::UserScope, "qmc2");
    qmc2Config = config;
  }

  QString userScopePath = QDir::homePath() + "/.qmc2";

  // Frontend

  // GUI
  checkBoxSaveLayout->setChecked(config->value("Frontend/GUI/SaveLayout", TRUE).toBool());
  checkBoxRestoreLayout->setChecked(config->value("Frontend/GUI/RestoreLayout", TRUE).toBool());
  checkBoxSaveGameSelection->setChecked(config->value("Frontend/GUI/SaveGameSelection", TRUE).toBool());
  checkBoxRestoreGameSelection->setChecked(config->value("Frontend/GUI/RestoreGameSelection", TRUE).toBool());
  checkBoxStatusbar->setChecked(config->value("Frontend/GUI/Statusbar", TRUE).toBool());
  checkBoxToolbar->setChecked(config->value("Frontend/GUI/Toolbar", TRUE).toBool());
  qmc2ScaledPreview = config->value("Frontend/GUI/ScaledPreview", TRUE).toBool();
  checkBoxScaledPreview->setChecked(qmc2ScaledPreview);
  qmc2ScaledFlyer = config->value("Frontend/GUI/ScaledFlyer", TRUE).toBool();
  checkBoxScaledFlyer->setChecked(qmc2ScaledFlyer);
  qmc2SmoothScaling = config->value("Frontend/GUI/SmoothScaling", FALSE).toBool();
  checkBoxSmoothScaling->setChecked(qmc2SmoothScaling);
  comboBoxLanguage->setCurrentIndex(comboBoxLanguage->findText(config->value("Frontend/GUI/Language", "us").toString().toUpper(), Qt::MatchContains | Qt::MatchCaseSensitive));
  comboBoxStyle->clear();
  comboBoxStyle->addItem(tr("Default"));
  comboBoxStyle->addItems(QStyleFactory::keys());
  QString myStyle = QObject::tr((const char *)config->value("Frontend/GUI/Style", tr("Default")).toString().toAscii());
  comboBoxStyle->setCurrentIndex(comboBoxStyle->findText(myStyle, Qt::MatchFixedString));
  lineEditFont->setText(config->value("Frontend/GUI/Font").toString());
  QFont f;
  f.fromString(lineEditFont->text());
  lineEditFont->setFont(f);
  int pixmapCacheSize = config->value("Frontend/GUI/PixmapCacheSize", 64).toInt();
  spinBoxPixmapCacheSize->setValue(pixmapCacheSize);
  checkBoxKillEmulatorsOnExit->setChecked(config->value("Frontend/GUI/KillEmulatorsOnExit", TRUE).toBool());
  
  // Files and directories
  lineEditDataDirectory->setText(config->value("Frontend/FilesAndDirectories/DataDirectory", "data/").toString());
  lineEditTemporaryFile->setText(config->value("Frontend/FilesAndDirectories/TemporaryFile", userScopePath + "/qmc2.tmp").toString());
  lineEditFrontendLogFile->setText(config->value("Frontend/FilesAndDirectories/LogFile", userScopePath + "/qmc2.log").toString());
  lineEditPreviewDirectory->setText(config->value("Frontend/FilesAndDirectories/PreviewDirectory", "data/prv/").toString());
  lineEditPreviewFile->setText(config->value("Frontend/FilesAndDirectories/PreviewFile", "data/prv/previews.zip").toString());
  qmc2UsePreviewFile = config->value("Frontend/FilesAndDirectories/UsePreviewFile", FALSE).toBool();
  stackedWidgetPreview->setCurrentIndex(qmc2UsePreviewFile ? 1 : 0);
  radioButtonPreviewSelect->setText(qmc2UsePreviewFile ? tr("Preview file") : tr("Preview directory"));
  lineEditFlyerDirectory->setText(config->value("Frontend/FilesAndDirectories/FlyerDirectory", "data/fly/").toString());
  lineEditFlyerFile->setText(config->value("Frontend/FilesAndDirectories/FlyerFile", "data/fly/flyers.zip").toString());
  qmc2UseFlyerFile = config->value("Frontend/FilesAndDirectories/UseFlyerFile", FALSE).toBool();
  stackedWidgetFlyer->setCurrentIndex(qmc2UseFlyerFile ? 1 : 0);
  radioButtonFlyerSelect->setText(qmc2UseFlyerFile ? tr("Flyer file") : tr("Flyer directory"));
  lineEditIconDirectory->setText(config->value("Frontend/FilesAndDirectories/IconDirectory", "data/ico/").toString());
  lineEditIconFile->setText(config->value("Frontend/FilesAndDirectories/IconFile", "data/ico/icons.zip").toString());
  qmc2UseIconFile = config->value("Frontend/FilesAndDirectories/UseIconFile", FALSE).toBool();
  stackedWidgetIcon->setCurrentIndex(qmc2UseIconFile ? 1 : 0);
  radioButtonIconSelect->setText(qmc2UseIconFile ? tr("Icon file") : tr("Icon directory"));
  lineEditOptionsTemplateFile->setText(config->value("Frontend/FilesAndDirectories/OptionsTemplateFile", "data/opt/template.xml").toString());
  lineEditFavoritesFile->setText(config->value("Frontend/FilesAndDirectories/FavoritesFile", userScopePath + "/qmc2.fav").toString());
  lineEditHistoryFile->setText(config->value("Frontend/FilesAndDirectories/HistoryFile", userScopePath + "/qmc2.hst").toString());
  lineEditROMStateCacheFile->setText(config->value("Frontend/FilesAndDirectories/ROMStateCacheFile", userScopePath + "/qmc2.rsc").toString());

  // Gamelist
  checkBoxSortOnline->setChecked(config->value("Frontend/Gamelist/SortOnline", FALSE).toBool());
  spinBoxResponsiveness->setValue(config->value("Frontend/Gamelist/Responsiveness", 100).toInt());
  qmc2GamelistResponsiveness = spinBoxResponsiveness->value();
  comboBoxSortCriteria->setCurrentIndex(config->value("Frontend/Gamelist/SortCriteria", 0).toInt());
  qmc2SortCriteria = comboBoxSortCriteria->currentIndex();
  comboBoxSortOrder->setCurrentIndex(config->value("Frontend/Gamelist/SortOrder", 0).toInt());
  qmc2SortOrder = comboBoxSortOrder->currentIndex() == 0 ? Qt::AscendingOrder : Qt::DescendingOrder;
  qmc2Filter.setBit(QMC2_ROMSTATE_INT_C, config->value("Frontend/Gamelist/ShowC", TRUE).toBool());
  toolButtonShowC->setChecked(qmc2Filter[QMC2_ROMSTATE_INT_C]);
  qmc2Filter.setBit(QMC2_ROMSTATE_INT_M, config->value("Frontend/Gamelist/ShowM", TRUE).toBool());
  toolButtonShowM->setChecked(qmc2Filter[QMC2_ROMSTATE_INT_M]);
  qmc2Filter.setBit(QMC2_ROMSTATE_INT_I, config->value("Frontend/Gamelist/ShowI", TRUE).toBool());
  toolButtonShowI->setChecked(qmc2Filter[QMC2_ROMSTATE_INT_I]);
  qmc2Filter.setBit(QMC2_ROMSTATE_INT_N, config->value("Frontend/Gamelist/ShowN", TRUE).toBool());
  toolButtonShowN->setChecked(qmc2Filter[QMC2_ROMSTATE_INT_N]);
  qmc2Filter.setBit(QMC2_ROMSTATE_INT_U, config->value("Frontend/Gamelist/ShowU", TRUE).toBool());
  toolButtonShowU->setChecked(qmc2Filter[QMC2_ROMSTATE_INT_U]);

  // MAME

  // Configuration
  if ( qmc2GuiReady )
    qmc2GlobalMAMEOptions->load();

  // Files and directories
  lineEditExecutableFile->setText(config->value("MAME/FilesAndDirectories/ExecutableFile", "").toString());
  lineEditMAMELogFile->setText(config->value("MAME/FilesAndDirectories/LogFile", userScopePath + "/mame.log").toString());

  if ( useDefaultSettings ) {
    QString fn = config->fileName();
    delete config;
    QFile f0(fn);
    f0.remove();
    QFile f(fn + ".bak");
    f.copy(fn);
    f.remove();
    config = new QSettings(QSettings::IniFormat, QSettings::UserScope, "qmc2");
    qmc2Config = config;
  }
}

void Options::on_toolButtonBrowseTemporaryFile_clicked()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::on_toolButtonBrowseTemporaryFile_clicked()");
#endif

  QFileInfo fileInfo(lineEditTemporaryFile->text());
  QString s = QFileDialog::getOpenFileName(this, tr("Choose temporary work file"), fileInfo.absolutePath(), tr("All files (*)"));
  if ( !s.isNull() )
    lineEditTemporaryFile->setText(s);
  raise();
}

void Options::on_toolButtonBrowsePreviewDirectory_clicked()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::on_toolButtonBrowsePreviewDirectory_clicked()");
#endif

  QString s = QFileDialog::getExistingDirectory(this, tr("Choose preview directory"), lineEditPreviewDirectory->text(), QFileDialog::ShowDirsOnly | QFileDialog::DontResolveSymlinks);
  if ( !s.isNull() )
    lineEditPreviewDirectory->setText(s);
  raise();
}

void Options::on_toolButtonBrowseFlyerDirectory_clicked()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::on_toolButtonBrowseFlyerDirectory_clicked()");
#endif

  QString s = QFileDialog::getExistingDirectory(this, tr("Choose flyer directory"), lineEditFlyerDirectory->text(), QFileDialog::ShowDirsOnly | QFileDialog::DontResolveSymlinks);
  if ( !s.isNull() )
    lineEditFlyerDirectory->setText(s);
  raise();
}

void Options::on_toolButtonBrowseIconDirectory_clicked()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::on_toolButtonBrowseIconDirectory_clicked()");
#endif

  QString s = QFileDialog::getExistingDirectory(this, tr("Choose icon directory"), lineEditIconDirectory->text(), QFileDialog::ShowDirsOnly | QFileDialog::DontResolveSymlinks);
  if ( !s.isNull() )
    lineEditIconDirectory->setText(s);
  raise();
}

void Options::on_toolButtonBrowseOptionsTemplateFile_clicked()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::on_toolButtonBrowseOptionsTemplateFile_clicked()");
#endif

  QFileInfo fileInfo(lineEditOptionsTemplateFile->text());
  QString s = QFileDialog::getOpenFileName(this, tr("Choose options template file"), fileInfo.absolutePath(), tr("All files (*)"));
  if ( !s.isNull() )
    lineEditOptionsTemplateFile->setText(s);
  raise();
}

void Options::on_toolButtonBrowseExecutableFile_clicked()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::on_toolButtonBrowseExecutableFile_clicked()");
#endif

  QFileInfo fileInfo(lineEditExecutableFile->text());
  QString s = QFileDialog::getOpenFileName(this, tr("Choose MAME executable file"), fileInfo.absolutePath(), tr("All files (*)"));
  if ( !s.isNull() )
    lineEditExecutableFile->setText(s);
  raise();
}

void Options::on_toolButtonBrowseMAMELogFile_clicked()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::on_toolButtonBrowseMAMELogFile_clicked()");
#endif

  QFileInfo fileInfo(lineEditMAMELogFile->text());
  QString s = QFileDialog::getOpenFileName(this, tr("Choose MAME log file"), fileInfo.absolutePath(), tr("All files (*)"));
  if ( !s.isNull() )
    lineEditMAMELogFile->setText(s);
  raise();
}

void Options::on_toolButtonBrowseFavoritesFile_clicked()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::on_toolButtonBrowseFavoritesFile_clicked()");
#endif

  QFileInfo fileInfo(lineEditFavoritesFile->text());
  QString s = QFileDialog::getOpenFileName(this, tr("Choose game favorites file"), fileInfo.absolutePath(), tr("All files (*)"));
  if ( !s.isNull() )
    lineEditFavoritesFile->setText(s);
  raise();
}

void Options::on_toolButtonBrowseHistoryFile_clicked()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::on_toolButtonBrowseHistoryFile_clicked()");
#endif

  QFileInfo fileInfo(lineEditHistoryFile->text());
  QString s = QFileDialog::getOpenFileName(this, tr("Choose play history file"), fileInfo.absolutePath(), tr("All files (*)"));
  if ( !s.isNull() )
    lineEditHistoryFile->setText(s);
  raise();
}

void Options::on_toolButtonBrowseROMStateCacheFile_clicked()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::on_toolButtonBrowseROMStateCacheFile_clicked()");
#endif

  QFileInfo fileInfo(lineEditROMStateCacheFile->text());
  QString s = QFileDialog::getOpenFileName(this, tr("Choose ROM state cache file"), fileInfo.absolutePath(), tr("All files (*)"));
  if ( !s.isNull() )
    lineEditROMStateCacheFile->setText(s);
  raise();
}

void Options::on_toolButtonBrowseFrontendLogFile_clicked()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::on_toolButtonBrowseFrontendLogFile_clicked()");
#endif

  QFileInfo fileInfo(lineEditFrontendLogFile->text());
  QString s = QFileDialog::getOpenFileName(this, tr("Choose frontend log file"), fileInfo.absolutePath(), tr("All files (*)"));
  if ( !s.isNull() )
    lineEditFrontendLogFile->setText(s);
  raise();
}

void Options::on_toolButtonBrowseDataDirectory_clicked()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::on_toolButtonBrowseDataDirectory_clicked()");
#endif

  QString s = QFileDialog::getExistingDirectory(this, tr("Choose data directory"), lineEditDataDirectory->text(), QFileDialog::ShowDirsOnly | QFileDialog::DontResolveSymlinks);
  if ( !s.isNull() )
    lineEditDataDirectory->setText(s);
  raise();
}

void Options::on_toolButtonBrowseFont_clicked()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::on_toolButtonBrowseFont_clicked()");
#endif

  bool ok;
  QFont currentFont;
  currentFont.fromString(lineEditFont->text());
  QFont f = QFontDialog::getFont(&ok, currentFont, 0);
  if ( ok ) {
    lineEditFont->setFont(f);
    lineEditFont->setText(f.toString());
  }
  raise();
}

void Options::closeEvent(QCloseEvent *e)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::closeEvent(QCloseEvent *e = 0x" + QString::number((ulong)e, 16) + ")");
#endif

  e->accept();
}

void Options::showEvent(QShowEvent *e)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::showEvent(QShowEvent *e = 0x" + QString::number((ulong)e, 16) + ")");
#endif

  if ( !qmc2CleaningUp && !qmc2EarlyStartup )
    if ( config->value("Frontend/GUI/SaveLayout").toBool() )
      config->setValue("Frontend/Layout/OptionsWidget/Visible", TRUE);

  e->accept();
}

void Options::hideEvent(QHideEvent *e)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::hideEvent(QHideEvent *e = 0x" + QString::number((ulong)e, 16) + ")");
#endif

  if ( !qmc2CleaningUp && !qmc2EarlyStartup )
    if ( config->value("Frontend/GUI/SaveLayout").toBool() )
      config->setValue("Frontend/Layout/OptionsWidget/Visible", FALSE);

  e->accept();
}

void Options::moveEvent(QMoveEvent *e)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::moveEvent(QMoveEvent *e = 0x" + QString::number((ulong)e, 16) + ")");
#endif

  if ( !qmc2CleaningUp && !qmc2EarlyStartup )
    if ( config->value("Frontend/GUI/SaveLayout").toBool() )
      config->setValue("Frontend/Layout/OptionsWidget/Position", pos());

  e->accept();
}

void Options::resizeEvent(QResizeEvent *e)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::resizeEvent(QResizeEvent *e = 0x" + QString::number((ulong)e, 16) + ")");
#endif

  if ( !qmc2CleaningUp && !qmc2EarlyStartup )
    if ( config->value("Frontend/GUI/SaveLayout").toBool() )
      config->setValue("Frontend/Layout/OptionsWidget/Size", size());

  e->accept();
}

void Options::on_radioButtonPreviewSelect_clicked()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::on_radioButtonPreviewSelect_clicked()");
#endif

  bool currentUsePreviewFile = (stackedWidgetPreview->currentIndex() == 1);
  stackedWidgetPreview->setCurrentIndex(!currentUsePreviewFile);
  radioButtonPreviewSelect->setText(!currentUsePreviewFile ? tr("Preview file") : tr("Preview directory"));
}

void Options::on_radioButtonFlyerSelect_clicked()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::on_radioButtonFlyerSelect_clicked()");
#endif

  bool currentUseFlyerFile = (stackedWidgetFlyer->currentIndex() == 1);
  stackedWidgetFlyer->setCurrentIndex(!currentUseFlyerFile);
  radioButtonFlyerSelect->setText(!currentUseFlyerFile ? tr("Flyer file") : tr("Flyer directory"));
}

void Options::on_radioButtonIconSelect_clicked()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::on_radioButtonIconSelect_clicked()");
#endif

  bool currentUseIconFile = (stackedWidgetIcon->currentIndex() == 1);
  stackedWidgetIcon->setCurrentIndex(!currentUseIconFile);
  radioButtonIconSelect->setText(!currentUseIconFile ? tr("Icon file") : tr("Icon directory"));
}

void Options::on_toolButtonBrowsePreviewFile_clicked()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::on_toolButtonBrowsePreviewFile_clicked()");
#endif

  QFileInfo fileInfo(lineEditPreviewFile->text());
  QString s = QFileDialog::getOpenFileName(this, tr("Choose ZIP-compressed preview file"), fileInfo.absolutePath(), tr("All files (*)"));
  if ( !s.isNull() )
    lineEditPreviewFile->setText(s);
  raise();
}

void Options::on_toolButtonBrowseFlyerFile_clicked()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::on_toolButtonBrowseFlyerFile_clicked()");
#endif

  QFileInfo fileInfo(lineEditFlyerFile->text());
  QString s = QFileDialog::getOpenFileName(this, tr("Choose ZIP-compressed flyer file"), fileInfo.absolutePath(), tr("All files (*)"));
  if ( !s.isNull() )
    lineEditFlyerFile->setText(s);
  raise();
}

void Options::on_toolButtonBrowseIconFile_clicked()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Options::on_toolButtonBrowseIconFile_clicked()");
#endif

  QFileInfo fileInfo(lineEditIconFile->text());
  QString s = QFileDialog::getOpenFileName(this, tr("Choose ZIP-compressed icon file"), fileInfo.absolutePath(), tr("All files (*)"));
  if ( !s.isNull() )
    lineEditIconFile->setText(s);
  raise();
}
