#include <QPixmapCache>
#include <QPixmap>
#include <QImage>
#include <QMatrix>
#include <QDir>
#include <QByteArray>

#include "preview.h"
#include "options.h"
#include "gamelist.h"
#include "qmc2main.h"
#include "macros.h"

// external global variables
extern MainWindow *qmc2MainWindow;
extern Options *qmc2Options;
extern Gamelist *qmc2Gamelist;
extern bool qmc2UsePreviewFile;
extern bool qmc2GuiReady;
extern bool qmc2ReloadActive;
extern bool qmc2ScaledPreview;
extern bool qmc2SmoothScaling;
extern QTreeWidgetItem *qmc2CurrentItem;
extern QSettings *qmc2Config;

QPixmap *currentPreviewPixmap;

Preview::Preview(QWidget *parent)
  : QFrame(parent)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Preview::Preview(QWidget *parent = 0x" + QString::number((ulong)parent, 16) + ")");
#endif

  setToolTip(tr("Game preview image"));
  setStatusTip(tr("Game preview image"));
  previewFile = NULL;
  if ( qmc2UsePreviewFile ) {
    previewFile = unzOpen((const char *)qmc2Config->value("Frontend/FilesAndDirectories/PreviewFile").toString().toAscii());
    if ( previewFile == NULL )
      qmc2MainWindow->log(LOG_QMC2, tr("FATAL: can't open preview file, please check access permissions for %1").arg(qmc2Config->value("Frontend/FilesAndDirectories/PreviewFile").toString()));
  }
}

Preview::~Preview()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Preview::~Preview()");
#endif

  if ( qmc2UsePreviewFile )
    unzClose(previewFile);
}

void Preview::paintEvent(QPaintEvent *e)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Preview::paintEvent(QPaintEvent *e = 0x" + QString::number((ulong)e, 16) + ")");
#endif

  if ( !qmc2CurrentItem )
    return;

  if ( qmc2CurrentItem->text(0) == tr("Waiting for data...") )
    return;

  QTreeWidgetItem *topLevelItem = qmc2CurrentItem;
  while ( topLevelItem->parent() )
    topLevelItem = topLevelItem->parent();

  QPainter p(this);

  QString gameName = topLevelItem->child(0)->text(1);
  static QPixmap cachedPixmap;

  if ( QPixmapCache::find(gameName, cachedPixmap) ) {
    currentPreviewPixmap = &cachedPixmap;
  } else {
    qmc2CurrentItem = topLevelItem;
    loadPreview(gameName);
  }

  if ( qmc2ScaledPreview )
    drawScaledImage(currentPreviewPixmap, &p);
  else
    drawCenteredImage(currentPreviewPixmap, &p);
}

bool Preview::loadPreview(QString gameName, bool checkOnly, QString *fileName)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, QString("DEBUG: Preview::loadPreview(QString gameName = %1, bool checkOnly = %2, QString *fileName = %3)").arg(gameName).arg(checkOnly).arg((qulonglong)fileName));
#endif

  static QPixmap pm;
  static char imageBuffer[QMC2_ZIP_BUFFER_SIZE];

  if ( fileName )
    *fileName = "";

  bool fileOk = TRUE;

  if ( qmc2UsePreviewFile ) {
    // use preview file
    QByteArray imageData;
    int len, i;
    QString gameFile = gameName + ".png";

    if ( fileName )
      *fileName = gameFile;

    if ( unzLocateFile(previewFile, (const char *)gameFile.toAscii(), 0) == UNZ_OK ) {
      if ( unzOpenCurrentFile(previewFile) == UNZ_OK ) {
        while ( (len = unzReadCurrentFile(previewFile, &imageBuffer, QMC2_ZIP_BUFFER_SIZE)) > 0 ) {
          for (i = 0; i < len; i++)
            imageData += imageBuffer[i];
        }
        unzCloseCurrentFile(previewFile);
      } else
        fileOk = FALSE;
    } else
      fileOk = FALSE;

    if ( fileOk )
      fileOk = pm.loadFromData(imageData);

    if ( !checkOnly ) {
      if ( fileOk ) {
        QPixmapCache::insert(gameName, pm); 
        currentPreviewPixmap = &pm;
      } else
        currentPreviewPixmap = &qmc2MainWindow->qmc2GhostImagePixmap;
    }
  } else {
    // use preview directory
    QString imageDir = qmc2Config->value("Frontend/FilesAndDirectories/PreviewDirectory").toString() + gameName;
    QString imagePath = imageDir + ".png";

    if ( fileName )
      *fileName = gameName + ".png";

    QFile f(imagePath);
    if ( !f.exists() ) {
      QDir dir(imageDir);
      if ( dir.exists() ) {
        QStringList nameFilter;
        nameFilter << "*.png";
        imagePath = imageDir + "/" + dir.entryList(nameFilter, QDir::Files | QDir::NoDotAndDotDot | QDir::Readable | QDir::CaseSensitive, QDir::Name | QDir::Reversed)[0];
        if ( fileName )
          *fileName = imagePath.remove(imageDir + "/");
      }
    }

    if ( checkOnly ) {
      fileOk = pm.load(imagePath);
    } else {
      if ( pm.load(imagePath) ) {
        QPixmapCache::insert(gameName, pm); 
        currentPreviewPixmap = &pm;
        fileOk = TRUE;
      } else {
        currentPreviewPixmap = &qmc2MainWindow->qmc2GhostImagePixmap;
        fileOk = FALSE;
      }
    }
  }

  return fileOk;
}

void Preview::drawCenteredImage(QPixmap *pm, QPainter *p)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Preview::drawCenteredImage(QPixmap *pm = 0x" + QString::number((ulong)pm, 16) + ", QPainter *p = 0x" + QString::number((ulong)p, 16) + ")");
#endif

  p->eraseRect(frameRect());

  if ( pm == NULL ) {
    p->end();
    return;
  }

  int posx = (frameRect().width() - pm->width()) / 2;
  int posy = (frameRect().height() - pm->height()) / 2;

  p->drawPixmap(posx, posy, *pm);
  p->end();
}

void Preview::drawScaledImage(QPixmap *pm, QPainter *p)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Preview::drawScaledImage(QPixmap *pm = 0x" + QString::number((ulong)pm, 16) + ", QPainter *p = 0x" + QString::number((ulong)p, 16) + ")");
#endif

  if ( pm == NULL ) {
    p->eraseRect(frameRect());
    p->end();
    return;
  }

  double desired_width;
  double desired_height;

  if ( pm->width() > pm->height() ) {
    desired_width  = contentsRect().width();
    desired_height = (double)pm->height() * (desired_width / (double)pm->width());
    if ( desired_height > contentsRect().height() ) {
      desired_height = contentsRect().height();
      desired_width  = (double)pm->width() * (desired_height / (double)pm->height());
    }
  } else {
    desired_height = contentsRect().height();
    desired_width  = (double)pm->width() * (desired_height / (double)pm->height());
    if ( desired_width > contentsRect().width() ) {
      desired_width = contentsRect().width();
      desired_height = (double)pm->height() * (desired_width / (double)pm->width());
    }
  }

  QPixmap pmScaled;

  if ( qmc2SmoothScaling )
    pmScaled = pm->scaled((int)desired_width, (int)desired_height, Qt::KeepAspectRatio, Qt::SmoothTransformation);
  else
    pmScaled = pm->scaled((int)desired_width, (int)desired_height, Qt::KeepAspectRatio, Qt::FastTransformation);

  drawCenteredImage(&pmScaled, p);
}
